/*
 * ex3_2_qpara.cc
 *
 * Validation code of blow-up solution of KS via Quasi-parabolic compactification
 *
 * Space compactification: Quasi-parabolic compactification
 * **Notice** this code is specialized for the FvKS
 *
 * written  Jul. 13, 2017 Akitoshi Takayasu
 *
 */
#include <kv/allsol.hpp>
#include <kv/ode-maffine.hpp>
#include <kv/kraw-approx.hpp>
// #include <kv/eig.hpp>
// #include "eig.hpp"
#include "eig1.hpp"
namespace ub = boost::numeric::ublas;
const double L = 1.0;
const int d = 3;
const int N = 6;
const double h = L/N;
ub::vector< kv::interval<double> > r(N-1); // r(0)=r_{1+1/2},...,r(N-2)=r_{N-1+1/2}: control volume
ub::vector< kv::interval<double> > r_pt(N); // r_pt(0)=r_1,...,r_pt(N-1)=r_N: control points

// const int size_of_ode = 2;    // Size of ODEs: fixed
const int k = 1; // quasi-homogeneous vector field of the order k+1: fixed
ub::vector<int> alph(2*N); // {alpha_i}: 2=alph_0=...=alph_{N-1}, 1=alph_N=...=alph_{2N-1};
ub::vector<int> beta(2*N); //{beta_i}: 1=beta_0=...=beta_{N-1}, 2=beta_N=...=beta_{2N-1};
const int gamm = 2; // c = alpha_1*beta_1 = ... = alpha_n*beta_n: fixed

ub::matrix< kv::interval<double> > Y;
ub::vector< kv::interval<double> > p;
ub::vector< kv::interval<double> > u0;
ub::vector< kv::interval<double> > py;


// Define tilde{f}(x): you must edit for each ODE.
struct Func_tilde {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
        T kappa;
        kappa = 0;
        for (int i = 0; i < 2*N; i++) {
          kappa += pow(x(i),2*beta(i));
        }
        kappa = 1.-kappa;
        // dy(0)=u_1, dy(1)=u_2,...,dy(N-1)=u_N,
        // dy(N)=v_1, dy(N+1)=v_2,...,dy(2*N-1)=v_N,
        // u_i, i=1,...,N
        ub::vector<T> dy(2*N);
        dy(0) = pow(r_pt(0),1-d)/h/h*pow(r(0),d-1)*((x(1)-x(0))*kappa-x(0)*(x(N+1)-x(N)));
        for (int i = 1; i < N-1; i++) {
          dy(i) = pow(r_pt(i),1-d)/h/h*(
            pow(r(i),d-1)*((x(i+1)-x(i))*kappa-(x(i+N+1)-x(i+N))*x(i))
            - pow(r(i-1),d-1)*((x(i)-x(i-1))*kappa-(x(i+N)-x(i+N-1))*x(i-1))
          );
        }
        dy(N-1) = pow(r_pt(N-1),1-d)/h/h*pow(r(N-2),d-1)*(-(x(N-1)-x(N-2))*kappa+(x(2*N-1)-x(2*N-2))*x(N-2));
        // v_i, i=1,...,N
        dy(N) = pow(r_pt(0),1-d)/h/h*pow(r(0),d-1)*(x(N+1)-x(N))*kappa - x(N)*kappa + x(0);
        for (int i = 1; i < N-1; i++) {
          dy(i+N) = pow(r_pt(i),1-d)/h/h*(
            pow(r(i),d-1)*(x(i+N+1)-x(i+N))
            -pow(r(i-1),d-1)*(x(i+N)-x(i+N-1))
          )*kappa - x(i+N)*kappa + x(i);
        }
        dy(2*N-1) = -pow(r_pt(N-1),1-d)/h/h*pow(r(N-2),d-1)*(x(2*N-1)-x(2*N-2))*kappa - x(2*N-1)*kappa + x(N-1);
        return dy;
    }
};
// Dominant term p_d(x): you must edit for each ODE.
struct Func_pd {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
      T kappa;
      kappa = 0;
      // dy(0)=u_1, dy(1)=u_2,...,dy(N-1)=u_N,
      // dy(N)=v_1, dy(N+1)=v_2,...,dy(2*N-1)=v_N,
      ub::vector<T> dy(2*N);
      dy(0) = pow(r_pt(0),1-d)/h/h*pow(r(0),d-1)*((x(1)-x(0))*kappa-x(0)*(x(N+1)-x(N)));
      for (int i = 1; i < N-1; i++) {
        dy(i) = pow(r_pt(i),1-d)/h/h*(
          pow(r(i),d-1)*((x(i+1)-x(i))*kappa-(x(i+N+1)-x(i+N))*x(i))
          - pow(r(i-1),d-1)*((x(i)-x(i-1))*kappa-(x(i+N)-x(i+N-1))*x(i-1))
        );
      }
      dy(N-1) = pow(r_pt(N-1),1-d)/h/h*pow(r(N-2),d-1)*(-(x(N-1)-x(N-2))*kappa+(x(2*N-1)-x(2*N-2))*x(N-2));
      // v_i, i=1,...,N
      dy(N) = pow(r_pt(0),1-d)/h/h*pow(r(0),d-1)*(x(N+1)-x(N))*kappa - x(N)*kappa + x(0);
      for (int i = 1; i < N-1; i++) {
        dy(i+N) = pow(r_pt(i),1-d)/h/h*(
          pow(r(i),d-1)*(x(i+N+1)-x(i+N))
          -pow(r(i-1),d-1)*(x(i+N)-x(i+N-1))
        )*kappa - x(i+N)*kappa + x(i);
      }
      dy(2*N-1) = -pow(r_pt(N-1),1-d)/h/h*pow(r(N-2),d-1)*(x(2*N-1)-x(2*N-2))*kappa - x(2*N-1)*kappa + x(N-1);
      return dy;
    }
};
// Generator of desingularized vector field
struct Func {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x, const T& t) {
      Func_tilde f;
      int n = 2*N;
      T kappa, tmp, G;
      kappa = 0;
      for (int i = 0; i < n; i++) {
        kappa += pow(x(i),2*beta(i));
      }
      tmp = 1-kappa;
      G = 0;
      for (int i = 0; i < N-1; i++) {
        G += 0.5/h/h*pow(r(i),d-1)*(
          -pow(r_pt(i+1),1-d)*x(i+1) + pow(r_pt(i),1-d)*x(i)
        )*(tmp*(x(i+1)-x(i))-(x(i+1+N)-x(i+N))*x(i)) + tmp/h/h*pow(r(i),d-1)*(
          -pow(r_pt(i+1),1-d)*pow(x(i+1+N),3) + pow(r_pt(i),1-d)*pow(x(i+N),3)
        )*(x(i+1+N)-x(i+N));
      }
      for (int i = 0; i < N; i++) {
        G += -tmp*pow(x(i+N),4)+pow(x(i+N),3)*x(i);
      }
      ub::vector<T> y(n), y_tmp(n);
      y_tmp = f(x);
      for (int i = 0; i < N; i++) {
        y(i) = 0.25*(1+3*kappa)*y_tmp(i) - 2*x(i)*G;
        y(i+N) = 0.25*(1+3*kappa)*y_tmp(i+N) - x(i+N)*G;
      }
      return y;
    }
};
// Func for kraw-approx
struct Func_krawapp {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
      Func_pd f;
      int n = 2*N;
      T kappa, tmp, G;
      kappa = 1;
      tmp = 0;
      G = 0;
      for (int i = 0; i < N-1; i++) {
        G += 0.5/h/h*pow(r(i),d-1)*(
          -pow(r_pt(i+1),1-d)*x(i+1) + pow(r_pt(i),1-d)*x(i)
        )*(tmp*(x(i+1)-x(i))-(x(i+1+N)-x(i+N))*x(i)) + tmp/h/h*pow(r(i),d-1)*(
          -pow(r_pt(i+1),1-d)*pow(x(i+1+N),3) + pow(r_pt(i),1-d)*pow(x(i+N),3)
        )*(x(i+1+N)-x(i+N));
      }
      for (int i = 0; i < N; i++) {
        G += -tmp*pow(x(i+N),4)+pow(x(i+N),3)*x(i);
      }
      ub::vector<T> y(n), y_tmp(n);
      y_tmp = f(x);
      for (int i = 0; i < N; i++) {
        y(i) = 0.25*(1+3*kappa)*y_tmp(i) - 2*x(i)*G;
        y(i+N) = 0.25*(1+3*kappa)*y_tmp(i+N) - x(i+N)*G;
      }
      return y;
    }
};
// Func for initial values
// struct Func_init {
//     template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
//       int n = 2*N;
//       T kappa;
//       kappa = 0;
//       for (int i = 0; i < n; i++) {
//         kappa += pow(x(i),2*beta(i));
//       }
//       kappa = 1 - kappa;
//       ub::vector<T> y(n);
//       for (int i = 0; i < n; i++) {
//         y(i) = pow(kappa,alph(i))*u0(i)-x(i);
//       }
//       return y;
//     }
// };
struct Func_init2 {
    template <class T> ub::vector<T> operator() (const ub::vector<T>& x) {
      // int n = 2*N;
      // T kappa;
      // kappa = 0;
      // for (int i = 0; i < n; i++) {
      //   kappa += pow(u0(i),2*beta(i));
      // }
      ub::vector<T> y(1);
      y(0) = pow(x(0),4)-pow(x(0),3)-py(0);
      return y;
    }
};
// ode_callback function:
// Integrate the time variable (t) and check the validated solution is in the Lyapunov domain.
namespace kv{
template <class T> struct ode_callback_sample : ode_callback<T> {
    interval<T> &result2;
    T eps;

    ode_callback_sample(interval<T>& result2, T eps) : result2(result2), eps(eps) {}

    virtual bool operator()(const interval<T>& start, const interval<T>& end, const ub::vector< interval<T> >& x_s, const ub::vector< interval<T> >& x_e, const ub::vector< psa< interval<T> > >& result) const {
        int i, size_of_result;
        ub::vector< interval<T> > xp;
        kv::interval<T> Lx;
        size_of_result = result.size();
        psa< interval<T> > px2c; // 1-p(x)^{2c}
        px2c = 1.;
        for (i = 0; i<size_of_result; i++) {
          px2c -= pow(result(i),2*beta(i));
        }
        if (k==1) {
          result2 += eval(integrate((1.0-0.5*(2*gamm-1)/gamm*px2c)*px2c),end-start);
        } else {
          result2 += eval(integrate((1.0-0.5*(2*gamm-1)/gamm*px2c)*pow(px2c,k)),end-start);
        }
        // Stopping criterion
        xp = x_e-p;
        Lx = inner_prod(prod(Y,xp),xp);
        if (Lx < pow(eps,(kv::interval<T>)2)) return false;
        return true;
    }
};
};

int main()
{
    int n = 2*N;    // # of unknowns
    // scalling order: c(=2) = alpha_1*beta_1 = ... = alpha_n*beta_n
    for (int i = 0; i < N; i++) {
      alph(i) = 2;
      alph(i+N) = 1;
      beta(i) = 1;
      beta(i+N) = 2;
    }
    // Control volumes & control points
    for (int i = 0; i < N-1; i++) {
      r(i) = i+1.;
      r_pt(i) = i+.5;
    }
    r_pt(N-1) = N-.5;
    //
    ub::vector< kv::interval<double> > x, xp, y0, x0, kappay;
    kv::interval<double> start, end, c1, cn_tmp, cn, Tmax, t_n, y, Lx, kappa;
    std::cout.precision(17);
    // u0: initial vector of u(t):
    u0.resize(n);
    for (int i = 0; i < N; i++) {
      // u0(i) = 10*exp(-2*pow(r_pt(i)*h,2)) + 20*exp(-2*pow(r_pt(i)*h-0.3,2));
      u0(i) = 100*(1+cos(kv::constants< kv::interval<double> >::pi()*r_pt(i)*h));
    }
    // std::cout << "u0: " << u0 << '\n';
    py.resize(1);
    py(0) = 0;
    for (int i = 0; i < n; i++) {
      py(0) += pow(u0(i),2*beta(i));
    }
    // Initial value: x(0)
    x0.resize(n);
    y0.resize(n);
    bool b = kv::krawczyk_approx(Func_init2(), mid(py), kappay, 100, 0);
    // std::cout << "kappa(y)" << kappay << '\n';
    // return 0;
    // bool b = true;
    for (int i = 0; i < n; i++) {
      x0(i) = u0(i)/pow(kappay(0),alph(i));
    }
    // std::cout << "x0: " << x0 << '\n';
    if (b) {
      // For check
      // for (int i = 0; i < N; i++) {
      //   x0(i+N) = 0;
      // }
      // std::cout << "x0: " << x0 << '\n';
      kappa = 0;
      for (int i = 0; i < n; i++) {
        kappa += pow(x0(i),2*beta(i));
      }
      kappa = 1 - kappa;
      // for (int i = 0; i < n; i++) {
      //   y0(i) = x0(i)/pow(kappa,alph(i));
      // }
      // std::cout << "y0: " << y0 << '\n';
      std::cout << "kappa: " << kappa << '\n';
    } else {
      std::cout << "Initial value is not found!" << '\n';
      return 1;
    }
    // std::cout << "alpha: " << alph << '\n';
    // std::cout << "beta: " << beta << '\n';
    //
    // Validation of the blow-up solution
    // 1. Solve the desingularized ode until some time and determine the critical point at inifity:
    // start = 0;
    // end = .05;
    while (true) {
      x = x0;
      // kv::odelong_maffine(Func(), x, start, end,c1,p,ub::identity_matrix< kv::interval<double> >(1));
      // kv::odelong_maffine(Func(), x, start, end, kv::ode_param<double>().set_verbose(0));
      x(0) = 0.99;
      x(1) = -0.036;
      x(N) = 0.13;
      x(N+1) = 0.02;
      p.resize(n);
      if (kv::krawczyk_approx(Func_krawapp(), mid(x), p, 15, 0)) {
        std::cout << "u_* = {" << '\n';
        for (int i = 0; i < N; i++) {
          std::cout << p(i) << '\n';
        }
        std::cout << "}" << '\n';
        std::cout << "v_* = {" << '\n';
        for (int i = 0; i < N; i++) {
          std::cout << p(i+N) << '\n';
        }
        std::cout << "}" << '\n';
        break;
      }
      end *= 1.5;
      if (end>5) {
        std::cout << "Cannot find equilibriums" << '\n';
        return 1;
      }
    }
    // std::cout << "end_time: " << end << '\n';
    // std::cout << "p: " << p << "\n";
    kappa = 1.;
    for (int i = 0; i < n; i++) {
      kappa -= pow(p(i),2*beta(i));
    }
    std::cout << "kappa: " << kappa << '\n';

    // 2. Search the maximum range of the Lyapunov domain:
    // Calculate C1
    Func f; //desingularized vector field of KS
    ub::vector< kv::interval<double> > y1, y2;
    ub::matrix< kv::interval<double> > Dfx;
    ub::vector< kv::autodif< kv::interval<double> > > dy1, dy2;
    // Initial vector of auto diff.
    y1.resize(n);
    for (int i = 0; i < n; i++) {
      y1(i) = p(i);
    }
    dy1 = kv::autodif< kv::interval<double> >::init(y1);
    dy2 = f(dy1,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
    kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
    // std::cout << "Df[p]: " << mid(Dfx) << std::endl;
    // Compute eigenvalue of Df(p)
    // Y=Q^T*Q s.t. Q^T*Dfx*Q = T (Shur decomposition)
    ub::vector< kv::complex< kv::interval<double> > > lambda;
    ub::matrix< kv::complex<double> > V, D;
    kv::eig(mid(Dfx),V,D);
    std::cout << "approximate eigenvalue:" << '\n';
    for (int i = 0; i < n; i++) {
      std::cout << D(i,i) << '\n';
      if (D(i,i).real()>0) {
        return 1;
      }
    }
    // return 0;
    // Shur decomposition is used for constructing Lyapunov function
    // ub::matrix<double> Q, H;
    // Q.resize(n,n);
    // H.resize(n,n);
    // kv::hess(mid(Dfx),Q,H);
    // kv::francisQR(Q,H);
    // Y = prod(trans(Q),Q);
    // Y = prod(Q,trans(Q));
    Y = ub::identity_matrix<double>(n);
    // std::cout << "Y: " << Y << std::endl;
    kv::veig(Y,lambda);
    // std::cout << "lambda_Y: " << lambda <<  std::endl;
    using std::abs;
    c1 = abs(lambda(0).real());
    for (int i = 1; i < n; i++) {
      if (c1 > abs(lambda(i).real())) {
        c1 = abs(lambda(i).real());
      }
    }
    c1 = 1./c1;
    std::cout << "c_1: " << c1 << std::endl;

    // Calculate Cn by the delta inflation scheme
    double eps_tmp = 1e-11, eps;// Radius of N
    // Func_krawapp fpd;
    ub::vector< kv::interval<double> > y1_tmp,radius;
    ub::matrix< kv::interval<double> > Ax, R;
    y1_tmp.resize(n);
    radius.resize(n);
    bool breakflg=false;
    while (true) {
      //------------------------- A(x) by using autodiff -----------------------
      for (int i = 0; i < n; i++) {
        radius(i) = eps_tmp*kv::interval<double>(-1.,1.);
      }
      y1_tmp = y1 + radius*sqrt(c1);
      // std::cout << "y1 = " << y1 << std::endl;
      dy1 = kv::autodif< kv::interval<double> >::init(y1_tmp);
      dy2 = f(dy1,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
      kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
  // ///////// Use higher derivatives for enclosing the Jacobian matrix /////////////
  //     ub::vector< kv::autodif< kv::autodif< kv::interval<double> > > > ddy1, ddy2;
  //     ub::matrix< kv::autodif< kv::interval<double> > > DF_I;
  //     ddy1 = kv::autodif< kv::autodif< kv::interval<double> > >::init(dy1);
  //     ddy2 = f(ddy1,(kv::autodif< kv::autodif< kv::interval<double> > >) 0);
  //     kv::autodif< kv::autodif< kv::interval<double> > >::split(ddy2, dy2, DF_I);
  //     // Autodiff at p: use fpd
  //     ub::matrix< kv::interval<double> > Dpd;
  //     ub::vector< kv::autodif< kv::interval<double> > > dy1_p, dy2_p;
  //     dy1_p = kv::autodif< kv::interval<double> >::init(y1);
  //     dy2_p = f(dy1_p,(kv::autodif< kv::interval<double> >) 0); // f(dy1) = [f(y1), df/dx[y1](Jacobian) (=Dfx)]
  //     dy2 = dy2_p + prod(DF_I,dy1-dy1_p);
  //     kv::autodif< kv::interval<double> >::split(dy2, y2, Dfx);
  //     // std::cout << "Dpd(p): " << Dpd << std::endl;
  //     // Autodiff around the neighborhood of p: use f
  //     // std::cout << "Df(I): " << Dfx << std::endl;
  // ////////////
      Ax = prod(trans(Dfx),Y)+prod(Y,Dfx);
      // R = trans(Dfx) + Dfx;
      // R = prod(trans(Q),R);
      // Ax = prod(R,Q);
      // std::cout << "Ax: " << Ax << std::endl;
      //------------------------- A(x) by using autodiff -----------------------
      kv::veig(Ax,lambda);
    std::cout << "lambda: " << '\n';
      for (int i = 0; i < n; i++) {
        std::cout << lambda(i).real() << '\n';
        if (lambda(i).real()>0) {
          // std::cout << "A(x) is not negative definite" << '\n';
          breakflg = true;
          break;
        }
      }
      // std::cout << "lambda_AX: " << lambda << std::endl;
      using std::abs;
      cn_tmp = abs(lambda(0));
      for (int i = 1; i < n; i++) {
        if (cn_tmp > abs(lambda(i))) {
          cn_tmp = abs(lambda(i));
        }
      }
      // std::cout << "c_n: " << cn << std::endl;
      // if (in(0,cn_tmp)) {
      if (in(0,cn_tmp)||rad(cn_tmp)>1e-1) { // Uncomment if you want the tight "cn".
        // return 0;
        breakflg = true;
      }
      if (breakflg) {
        if (eps_tmp==1e-10) {
          std::cout << "Failed in constructing the Lyapunov function" << '\n';
          return 1;
        }
        break;
      }
      y1 = y1_tmp;
      cn = cn_tmp;
      eps = eps_tmp;
      eps_tmp *= 1.1;
    }
    std::cout << "The Lyapunov domain is validated." << '\n';
    std::cout << "c_n: " << cn << '\n';
    std::cout << "eps = " << eps << '\n';
    //
    // 3. Validate that x(tau_N) is in the Lyapunov domain
    t_n  = 0; // Initialize the blow-up time
    x = x0;
    start = 0;
    end = 100;
    // std::cout << "t: " << t_n << std::endl;
    // std::cout << x << std::endl;
    int r = kv::odelong_maffine(
      Func(),
      x,
      start,
      end,
      // (kv::interval<double>)eps,
      // p,
      // Y,
      kv::ode_param<double>().set_verbose(0).set_order(24).set_restart_max(10).set_epsilon(pow(2,-68)),
      kv::ode_callback_sample<double>(t_n,eps)
    );
    // int r = kv::odelong_maffine0(Func(), x, start, end, kv::ode_param<double>().set_verbose(0).set_order(24).set_restart_max(1),kv::ode_callback_sample<double>(t_n));
    if (r == 0) {
        std::cout << "Cannot calculate solution.\n";
        return 1;
    } else if (r == 1 || r == 3) {
        std::cout << "Solution calculated until t = " << end << ".\n";
        std::cout << x << "\n";
    } else {
        // std::cout << "Solution completely calculated until t = " << end << ".\n";
        // std::cout << "x_n: " << x << "\n";
        std::cout << "Cannot valitate that x(tau_N) is in the Lyapunov domain." << '\n';
        return 1;
    }
    // std::cout << "Norm(x)" << norm_2(x) << std::endl;
    std::cout << "T_n: " << t_n << std::endl;
    xp = x-p;
    Lx = inner_prod(prod(Y,xp),xp);
    std::cout << "L(x): " << Lx << std::endl;

    // 4. Compute the validated blow-up time of the ODE
    //----------------- This part is specialized for 1-p(x)^4. -----------------
    // v = (...,2u_j^*,...,4(v_j^*)^3,...)
    // v_norm = |v|_2
    ub::vector< kv::interval<double> > v;
    kv::interval<double> v_norm;
    v.resize(n);
    for (int i = 0; i < N; i++) {
      v(i) = 2*p(i);
      v(i+N) = 4*pow(p(i+N),3);
    }
    v_norm = norm_2(v);
    // kappa = max{1, 6|v^*_j|^2}
    kv::interval<double> kappa_tmp;
    kappa = 0;
    for (int i = 0; i < N; i++) {
      kappa_tmp = 6*pow(abs(p(i+N)),2);
      if (1.0>kappa_tmp) {
        kappa_tmp = 1.0;
      }
      if (kappa_tmp>kappa) {
        kappa = kappa_tmp;
      }
    }
    // kappa3 = max{4|v_j|}
    kv::interval<double> kappa3;
    kappa3 = 0;
    for (int i = 0; i < N; i++) {
      kappa_tmp = 4*abs(p(i+N));
      if (kappa_tmp>kappa3){
        kappa3 = kappa_tmp;
      }
    }
    std::cout << "T_max-T_n: " << mag((
      2*v_norm*pow(Lx/c1,0.5)
      +kappa*Lx
      +2*kappa3*pow(c1,0.5)*pow(Lx,1.5)/3.0
      +0.5*c1*pow(Lx,2))/cn) << std::endl;
    Tmax = kv::interval<double>::hull(t_n,t_n+mag((
      2*v_norm*pow(Lx/c1,0.5)
      +kappa*Lx
      +2*kappa3*pow(c1,0.5)*pow(Lx,1.5)/3.0
      +0.5*c1*pow(Lx,2))/cn));
    //--------------------------------------------------------------------------
    std::cout << "Tmax: " << Tmax << std::endl;
}
